import logging
from typing import TYPE_CHECKING, List, Dict, Optional
from matching.rules import rule as rl

if TYPE_CHECKING:
    from matching.mentor import Mentor
    from matching.mentee import Mentee


class Match:
    """
    This is the class that calculates the score of each Match.

    As of v4, it requires a list of `Rule` objects that will
    determine the score. This object has two Rules built in: one to disqualify mentors and mentees being matched when
    they're the same people (as people may sign up as both), and one that disqualifies if they've previously been
    matched
    """

    def __init__(
        self,
        mentor: "Mentor",
        mentee: "Mentee",
        weightings: Dict[str, int],
        rules: Optional[List["rl.Rule"]],
    ):
        self.weightings = weightings
        self.mentee = mentee
        self.mentor = mentor
        self._disallowed: bool = False
        self._score: int = 0
        self.rules = [
            rl.Disqualify(lambda match: match.mentor.email == match.mentee.email),
            rl.Disqualify(
                lambda match: match.mentor in match.mentee.mentors
                or match.mentee in match.mentor.mentees
            ),
        ]
        if rules:
            self.rules.extend(rules)

    @property
    def score(self):
        if self._disallowed:
            return 0
        else:
            return self._score

    @score.setter
    def score(self, new_value: int):
        self._score += new_value

    @property
    def disallowed(self):
        return self._disallowed

    @disallowed.setter
    def disallowed(self, new_value: bool):
        if self._disallowed is False and new_value is True:
            self._disallowed = new_value

    def calculate_match(self) -> "Match":
        """
        This method calculates the score for this Match object. It does this by applying the functions below. If at any
        point the match becomes disallowed, the loop breaks. Note that if the match is disallowed, the score property
        always returns 0. One complete, it returns the object
        """
        while not self.disallowed and self.rules:
            rule = self.rules.pop()
            self.score += rule.apply(self)
        return self

    def mark_successful(self):
        if not self.disallowed:
            self.mentor.mentees.append(self.mentee)
            self.mentee.mentors.append(self.mentor)
        else:
            logging.debug("Skipping this match as disallowed")



import argparse
import logging
from pathlib import Path

from matching.process import conduct_matching_from_file, create_mailing_list


def main():
    parser = argparse.ArgumentParser(
        description='Match mentors to mentees. Your files should be called "mentees.csv" and "mentors.csv".'
    )
    parser.add_argument(
        "filepath", type=str, help="the path to the data containing the files"
    )
    path_to_data = Path(parser.parse_args().filepath)
    logging.info("Beginning matching exercise. This might take up to five minutes.")
    mentors, mentees = conduct_matching_from_file(path_to_data)
    logging.info("Matches found. Exporting to output folder!")
    out_put_folder = path_to_data / "output"
    create_mailing_list(mentors, out_put_folder)
    create_mailing_list(mentees, out_put_folder)


if __name__ == "__main__":
    main()



from matching.person import Person
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from matching.mentee import Mentee


class Mentor(Person):
    def __init__(self, **kwargs):
        super(Mentor, self).__init__(**kwargs)

    @property
    def mentees(self):
        return super(Mentor, self).connections

    @mentees.setter
    def mentees(self, new_mentee: "Mentee"):
        super(Mentor, self).connections.append(new_mentee)

    @classmethod
    def __str__(cls):
        return "mentor"



from typing import TYPE_CHECKING, Dict, Union, List
from matching.person import Person

if TYPE_CHECKING:
    from matching.mentor import Mentor


class Mentee(Person):
    def __init__(self, **kwargs):
        """
        Mentees should have a target profession. If they don't have a target profession, we assume it's the same one
        they're in at the moment
        :param kwargs:
        """
        self.target_profession = kwargs.get(
            "target profession", kwargs.get("current profession")
        )
        super(Mentee, self).__init__(**kwargs)

    @property
    def mentors(self):
        return super(Mentee, self).connections

    @mentors.setter
    def mentors(self, new_mentor: "Mentor"):
        super(Mentee, self).connections.append(new_mentor)

    def core_to_dict(self) -> Dict[str, Dict[str, Union[str, List]]]:
        core = super(Mentee, self).core_to_dict()
        core[self.class_name()]["target profession"] = self.target_profession
        return core

    @classmethod
    def __str__(cls):
        return "mentee"



import operator
from abc import ABC, abstractmethod
from typing import Callable, Dict, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from matching.match import Match


class AbstractRule(ABC):
    @abstractmethod
    def apply(self, match_object: "Match") -> int:
        raise NotImplementedError


class Rule(AbstractRule):
    def __init__(self, score_dict: Union[Dict[bool, int], None] = None):
        if score_dict is None:
            score_dict = {True: 0, False: 0}
        self.results = score_dict

    def apply(self, match_object: "Match") -> int:
        return self.results.get(self.evaluate(match_object), False)

    @abstractmethod
    def evaluate(self, match_object: "Match") -> bool:
        raise NotImplementedError


class UnmatchedBonus(Rule):
    def __init__(self, unmatched_bonus: int):
        super(UnmatchedBonus, self).__init__({True: unmatched_bonus, False: 0})

    def evaluate(self, match_object: "Match") -> bool:
        return any(
            map(
                lambda participant: len(participant.connections) == 0,
                (match_object.mentee, match_object.mentor),
            )
        )


class Grade(Rule):
    def __init__(
        self,
        target_diff: int,
        logical_operator: Callable[[int, int], bool],
        score_dict: Union[Dict[bool, int], None] = None,
    ):
        super(Grade, self).__init__(score_dict)
        self.target_diff = target_diff
        self.operator = logical_operator

    def evaluate(self, match_object: "Match") -> bool:
        return self.operator(
            (match_object.mentor.grade - match_object.mentee.grade), self.target_diff
        )


class Equivalent(Rule):
    def __init__(self, attribute: str, score_dict: Union[Dict[bool, int], None] = None):
        super(Equivalent, self).__init__(score_dict)
        self.attribute = attribute

    def evaluate(self, match_object: "Match") -> bool:
        participants = (match_object.mentor, match_object.mentee)
        attrs = map(
            lambda participant: participant.__getattribute__(self.attribute),
            participants,
        )
        return operator.eq(*attrs)


class Generic(Rule):
    def __init__(
        self,
        score_dict: Union[Dict[bool, int], None],
        evaluation_func: Callable[["Match"], bool],
    ):
        super(Generic, self).__init__(score_dict)
        self._evaluate = evaluation_func

    def evaluate(self, match_object: "Match") -> bool:
        return self._evaluate(match_object)


class Disqualify(Generic):
    """
    A disqualifying rule is a kind of anti-rule. Here, we pass a condition which, if it evaluates to `True`, should
    disqualify a `Match` rather than increase its score.
    """

    def __init__(self, disqualifying_condition: Callable[["Match"], bool]):
        super(Disqualify, self).__init__(None, disqualifying_condition)

    def apply(self, match_object: "Match") -> int:
        if super(Disqualify, self).evaluate(match_object):
            match_object.disallowed = True
        return 0



from typing import Dict, List

from matching.mentee import Mentee
from matching.mentor import Mentor
from matching.person import Person


class ParticipantFactory:
    participant_types = {"mentor": Mentor, "mentee": Mentee}

    @classmethod
    def create_from_dict(cls, data_as_dict: Dict):
        participant_type_str = list(data_as_dict).pop()
        participant_type = cls.participant_types.get(participant_type_str, Person)
        participant_data = data_as_dict.get(participant_type_str, dict())
        participant = participant_type(**participant_data)
        connections: List[Dict[str, str]] = participant_data.get("connections", [])
        participant._connections = [
            cls.create_from_dict(connection_data) for connection_data in connections
        ]
        if type(participant) is Mentee:
            participant.target_profession = participant_data["target profession"]
        return participant



import csv
import functools
import operator
import os
import pathlib
import sys
from pathlib import Path
from typing import Union, Type, List, Dict, Tuple, Generator, Optional, Callable

from munkres import Munkres, make_cost_matrix, Matrix  # type: ignore

from matching.match import Match
from matching.mentee import Mentee
from matching.mentor import Mentor
import matching.rules.rule as rl


def generate_match_matrix(
    mentor_list: List[Mentor],
    mentee_list: List[Mentee],
    weightings: Dict[str, int],
    rules: Optional[List[rl.Rule]] = None,
) -> List[List[Match]]:
    return [
        [
            Match(mentor, mentee, weightings, rules).calculate_match()
            for mentee in mentee_list
        ]
        for mentor in mentor_list
    ]


def process_form(path_to_form) -> Generator[Dict[str, str], None, None]:
    with open(path_to_form, "r") as data_form:
        file_reader = csv.DictReader(data_form)
        for row in file_reader:
            yield row


def create_participant_list_from_path(
    participant: Union[Type[Mentee], Type[Mentor]],
    path_to_data: pathlib.Path,
    mapping_func: Optional[
        Callable[[dict[str, str], str], dict[str, str]]
    ] = lambda row, name: row,
):
    path_to_data = path_to_data / f"{participant.__str__()}s.csv"
    return [
        participant(**mapping_func(row, participant.__str__()))
        for row in process_form(path_to_data)
    ]


def transpose_matrix(matrix):
    return [list(row) for row in zip(*matrix)]


def create_matches(
    preliminary_matches: List[List[Match]],
) -> List[List[Match]]:
    """
    Create a preliminary grid of matches and then strip out folks who have zero potential matches.
    :param: preliminary_matches:
    :return:
    """

    def _at_least_one_match(row: List[Match]) -> bool:
        return not all(match.disallowed for match in row)

    good_mentees = [row for row in preliminary_matches if _at_least_one_match(row)]
    good_mentors = [
        row for row in transpose_matrix(good_mentees) if _at_least_one_match(row)
    ]
    return transpose_matrix(good_mentors)


def prepare_matrix(matches: List[List[Match]]) -> Matrix:
    prepared_matrix = make_cost_matrix(
        matches,
        lambda match: sys.maxsize - match.score,
    )
    return prepared_matrix


def calculate_matches(prepared_matrix: Matrix):
    algorithm = Munkres()
    return algorithm.compute(prepared_matrix)


def match_and_assign_participants(
    good_matches: List[List[Match]],
) -> List[List[Match]]:
    for successful_match in calculate_matches(prepare_matrix(good_matches)):
        match = good_matches[successful_match[0]][successful_match[1]]
        match.mark_successful()
    return good_matches


def process_data(
    mentors: List[Mentor],
    mentees: List[Mentee],
    weightings_list: List[Dict[str, int]],
    all_rules: Optional[List[List[rl.Rule]]] = None,
) -> Tuple[List[Mentor], List[Mentee]]:
    """
    This is the main entrypoint for this software. It lazily generates three matrices, which allows for them to be
    mutated over the course of the matching process.
    :param all_rules:
    :param mentors:
    :param mentees:
    :param weightings_list:
    :return:
    """
    if not all_rules:
        base_rules: List[rl.AbstractRule] = [
            rl.Disqualify(
                lambda match: match.mentee.organisation == match.mentor.organisation
            ),
            rl.Disqualify(
                rl.Grade(target_diff=2, logical_operator=operator.gt).evaluate
            ),
            rl.Disqualify(
                rl.Grade(target_diff=0, logical_operator=operator.le).evaluate
            ),
        ]
        all_rules = [base_rules for _ in range(len(weightings_list))]
        for i, match_round in enumerate(weightings_list):
            unique_rules = [
                rl.Grade(1, operator.eq, {True: match_round.get("grade", 3), False: 0}),
                rl.Grade(2, operator.eq, {True: match_round.get("grade", 6), False: 0}),
                rl.Generic(
                    {True: match_round.get("profession", 0), False: 0},
                    lambda match: match.mentee.target_profession
                    == match.mentor.current_profession,
                ),
                rl.UnmatchedBonus(match_round.get("unmatched bonus", 0)),
            ]
            all_rules[i].extend(unique_rules)
    matrices = map(
        functools.partial(generate_match_matrix, mentors, mentees),
        weightings_list,
        all_rules,
    )
    for matrix in matrices:
        match_and_assign_participants(matrix)
    return mentors, mentees


def conduct_matching_from_file(
    path_to_data: Path, weightings_list: [List[Dict[str, int]]]
) -> Tuple[List[Mentor], List[Mentee]]:
    mentors = create_participant_list_from_path(Mentor, path_to_data)
    mentees = create_participant_list_from_path(Mentee, path_to_data)
    return process_data(mentors, mentees, weightings_list)


def create_mailing_list(
    participant_list: List[Union[Mentor, Mentee]], output_folder: Path
):
    """
    This function takes a list of either matched mentors or matched mentees. For each participant, it outputs their
    data and the information of the participants they've been matched with. If a participant doesn't have the full
    complement of three matches, the empty spaces are ignored.
    """
    file_name = f"{type(participant_list[0]).__str__()}s-list.csv"
    file = output_folder.joinpath(file_name)
    list_participants_as_dicts = [
        participant.to_dict_for_export() for participant in participant_list
    ]
    field_headings = max(
        list_participants_as_dicts, key=lambda participant: len(participant.keys())
    ).keys()
    try:
        os.mkdir(output_folder)
    except FileExistsError:
        pass
    with open(file, "w", newline="") as output_file:
        writer = csv.DictWriter(output_file, fieldnames=list(field_headings))
        writer.writeheader()
        for participant in list_participants_as_dicts:
            writer.writerow(participant)



from typing import List, Dict, Union


class Person:
    def __init__(self, **kwargs):
        """
        When creating a person from a dictionary, we expect a grade as an integer. The lower the `int`, the lower the
        grade. It is the client's responsibility to turn this integer back into a human-readable `str` if needed
        :param kwargs:
        """
        self.grade: int = int(kwargs.get("grade"))
        self.organisation: str = kwargs.get("organisation", None)
        self.current_profession: str = kwargs.get("current profession", None)
        self.email = kwargs.get("email", None)
        self.first_name = kwargs.get("first name", None)
        self.last_name = kwargs.get("last name", None)
        self.role = kwargs.get("role", None)
        self._connections: List[Person] = []
        self.has_no_match: bool = False

    @property
    def connections(self) -> List["Person"]:
        return self._connections

    @connections.setter
    def connections(self, new_connection: "Person"):
        if len(self._connections) < 3:
            self._connections.append(new_connection)
        else:
            raise Exception

    def to_dict(
        self,
    ) -> Dict[str, Dict[str, Union[str, List[Dict[str, Dict[str, str]]]]]]:
        output = self.core_to_dict()
        output[self.class_name()]["connections"] = [
            connection.core_to_dict() for connection in self.connections
        ]
        return output

    def core_to_dict(self) -> Dict[str, Dict[str, Union[str, List]]]:
        return {
            self.class_name(): {
                "email": self.email,
                "first name": self.first_name,
                "last name": self.last_name,
                "role": self.role,
                "organisation": self.organisation,
                "grade": self.grade,
                "current profession": self.current_profession,
            }
        }

    def to_dict_for_output(self, depth=1) -> dict:
        output = self.core_to_dict()[self.class_name()]
        if depth == 1:
            for i, connection in enumerate(self._connections):
                for key, value in connection.to_dict_for_output(depth=0).items():
                    output[f"match {i + 1} {key}"] = value
        return output

    def to_dict_for_export(self, depth=1) -> dict:
        """
        This method converts the class into a dictionary suitable for export. It might be useful if you don't want
        every attribute on the model exposed, or if there's some mapping you need to do - for example, 'grade' might be
        'rank' in your organisation. By default, this method just calls `to_dict_for_output`
        """
        return self.to_dict_for_output(depth)

    def class_name(self):
        return self.__class__.__name__.lower()

    def __eq__(self, other: "Person"):
        return self.email == other.email



import pathlib

from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="mentor-match",
    version="5.1.4",
    description="A series of classes to match mentors and mentees",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Jonathan Kerr",
    author_email="jonathan.drew.kerr@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["matching", "matching/rules"],
    include_package_data=True,
    install_requires=["munkres"],
    setup_requires=["wheel"],
)
